@@ -5,9 +5,9 @@
 #
 #   config.add_extension('_name', sources=['source_fname'], **numpy_nodepr_api)
 #
-numpy_nodepr_api = dict(
-    define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_9_API_VERSION")]
-)
+numpy_nodepr_api = {
+    "define_macros": [("NPY_NO_DEPRECATED_API", "NPY_1_9_API_VERSION")]
+}
 
 
 def import_file(folder, module_name):
